/*
 * config.c --- Parse a configuration file and build a key,value structure.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>
#include <assert.h>

#ifndef CFG_MAIN

#include "main.h"
#include "tonegen.h"
#include "midiIn.h"
#include "reverb.h"

#endif /* CFG_MAIN */

#include "config.h"

#define LINEBUFSZ 2048

/*
 * Each configurable module implements this function. The implementation
 * is idempotent. The most recent call defines the parameter's value.
 */
static int distributeParameter (ConfigContext * cfg) {


#ifdef CFG_MAIN

  printf ("%s:%d:[%s]=[%s]\n",
	  cfg->fname,
	  cfg->linenr,
	  cfg->name,
 	  cfg->value);

#else

  genConfig (cfg);
  midiConfig (cfg);
  oscConfig (cfg);
  // scnConfig (cfg);
  ampConfig (cfg);
  reverbConfig (cfg);
  rotConfig (cfg);

#endif /* CFG_MAIN */

  return 0;
}

/*
 *
 */
int parseConfigurationLine (char * fname, int lineNumber, char * oneLine) {
  char delim[] = "=\n";
  char * s = oneLine;
  char * name;
  char * value;

  while (isspace (*s)) s++;	/* Skip over leading spaces */
  if (*s == '\0') return 0;	/* Skip empty lines */
  if (*s == '#') return 0;	/* Skip comment lines */
  if (*s == '=') {
    fprintf (stderr,
	     "%s:line %d: empty parameter name.\n",
	     fname,
	     lineNumber);
    return 0;
  }

  if ((name = strtok (s, delim)) != NULL) {
    int i;

	for (i = strlen (name) - 1; isspace (name[i]); name[i] = '\0', i--);

	if ((value = strtok (NULL, delim)) != NULL) {
	  char * t;
	  while (isspace (value[0])) value++;
	  for (t = value; *t != '\0'; t++) {
	    if (*t == '#') {	/* Terminate value at comment */
	      *t = '\0';
	      break;
	    }
	  }
	}
	else {
	  value = "";
	}

	i = strlen (value);
	if (0 < i) {
	  for (i = i - 1;  isspace (value[i]); value[i] = '\0', i--);
	}

	if (strcasecmp (name, "config.read") == 0) {
	  parseConfigurationFile (value);
	}
	else {
	  ConfigContext cfg;
	  cfg.fname  = fname;
	  cfg.linenr = lineNumber;
	  cfg.name   = name;
	  cfg.value  = value;
	  distributeParameter (& cfg);
	}
      }

}

/*
 *
 */
int parseConfigurationFile (char * fname) {
  int lineNumber = 0;
  char lineBuf [LINEBUFSZ];
  FILE * fp;

  if ((fp = fopen (fname, "r")) == NULL) {
    perror (fname);
    return -1;
  }
  else {

    while (fgets (lineBuf, LINEBUFSZ, fp) != NULL) {
      lineNumber += 1;		/* Increment the linenumber. */
      parseConfigurationLine (fname, lineNumber, lineBuf);
    }

    fclose (fp);
  }

  return 0;
}

/*
 *
 */
void configIntUnparsable (ConfigContext * cfg) {
  fprintf (stderr,
	   "%s:line %d:%s:value is not an integer:%s\n",
	   cfg->fname,
	   cfg->linenr,
	   cfg->name,
	   cfg->value);
}

/*
 *
 */
void configDoubleUnparsable (ConfigContext * cfg) {
  fprintf (stderr,
	   "%s:line %d:%s:value is not a real:%s\n",
	   cfg->fname,
	   cfg->linenr,
	   cfg->name,
	   cfg->value);
}

/*
 *
 */
void configIntOutOfRange (ConfigContext * cfg, int lowest, int highest) {
  assert (cfg != NULL);
  fprintf (stderr,
	   "%s:line %d:%s value out range, min=%d, max=%d:%s\n",
	   cfg->fname,
	   cfg->linenr,
	   cfg->name,
	   lowest,
	   highest,
	   cfg->value);
}

/*
 * Assigns an sample range integer from the configuration context's value.
 * If the value is real (contains a period), it is interpreted as a
 * normalised quantity (-1.0 -- 1.0) and multiplied with 32767.0.
 * Otherwise the value is interpreted to be an integer and is assigned
 * as it is.
 */
void setConfigRangeInt (int * vp, ConfigContext * cfg) {

  assert (vp != NULL);
  assert (cfg != NULL);

  if (strchr (cfg->value, '.') != NULL) {
    double d;
    if (sscanf (cfg->value, "%lf", &d) == 1) {
      *vp = (int) (32767.0 * d);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else {
    int k;
    if (sscanf (cfg->value, "%d", &k) == 1) {
      *vp = k;
    }
    else {
      configIntUnparsable (cfg);
    }
  }

}

/*
 * Assigns an integer from the configuration context's value.
 */
void setConfigInt (int * vp, ConfigContext * cfg) {
  int k;
  assert (vp != NULL);
  assert (cfg != NULL);
  if (sscanf (cfg->value, "%d", &k) == 1) {
    *vp = k;
  }
  else {
    configIntUnparsable (cfg);
  }
}

/*
 * Assigns a double from the configuration context's value.
 */
void setConfigDouble (double * vp, ConfigContext * cfg) {
  double d;
  assert (vp != NULL);
  assert (cfg != NULL);
  if (sscanf (cfg->value, "%lf", &d) == 1) {
    *vp = d;
  }
  else {
    configDoubleUnparsable (cfg);
  }
}

/*
 * Returns the configuration entry's value string.
 */
char * getConfigValue (ConfigContext * cfg) {
  return cfg->value;
}

#ifdef CFG_MAIN
int main () {
  parseConfigurationFile ("test.cfg");
  return 0;
}
#endif /* CFG_MAIN */
