/*
 * midiIn.c
 * 2001-12-26/FK
 * MIDI input routines for Beatrix.
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/soundcard.h>
#include <limits.h>

#include "config.h"
#include "tonegen.h"
#include "preamp.h"
#include "rotsim.h"
#include "scanner.h"
#include "program.h"

#define BUFFER_SIZE_BYTES 1024

#define CONSUME_ONE_BYTE  5
#define CONSUME_TWO_BYTES 6

static int midi_fd;
static char * midiDevice = "/dev/midi00";
static char midiDeviceBuffer[PATH_MAX + 1];
static unsigned char rcvCh = 0;	/* MIDI receive channel */
static unsigned char inBuffer [BUFFER_SIZE_BYTES];

static char nsh = 0;

/*
 *
 */
int prepareMIDI () {
  int timeToWait = 1;
  int version;

  if ((midi_fd = open (midiDevice, O_RDONLY, 0)) == -1) {
    perror (midiDevice);
    exit (1);
  }

#ifdef COMMENT /* This worked under Red Hat 6.2 but not under Debian */
  if (ioctl (midi_fd, SNDCTL_MIDI_PRETIME, & timeToWait) == -1) {
    perror ("SNDCTL_MIDI_PRETIME");
    exit (1);
  }

  if (ioctl (midi_fd, OSS_GETVERSION, & version) == -1) {
    perror ("OSS_GETVERSION");
    exit (1);
  }

  fprintf (stderr, "MIDI DRIVER VERSION %x\n", version);
#endif /* COMMENT */

  return 0;
}

/*
 *
 */
void setMIDIDevice (char * s) {
  if (s != NULL) {
    strncpy (midiDeviceBuffer, s, PATH_MAX);
    midiDevice = midiDeviceBuffer;
  }
}

/*
 *
 */
void setMIDIChannelSwell (int ch) {
  rcvCh = ch & 0x000F;
}

/*
 *
 */
void setMIDINoteShift (char offset) {
  nsh = offset;
}

/*
 *
 */
int midiConfig (ConfigContext * cfg) {
  int v;
  if (strcasecmp (cfg->name, "midi.swell.channel") == 0) {
    if (sscanf (cfg->value, "%d", &v) == 1) {
      if ((1 <= v) && (v <= 16)) {
	setMIDIChannelSwell (v - 1);
      }
      else {
	configIntOutOfRange (cfg, 1, 16);
      }
    }
    else {
      configIntUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "midi.noteshift") == 0) {
    int v;
    if (sscanf (cfg->value, "%d", &v) == 1) {
      if ((-127 <= v) && (v <= 127)) {
	setMIDINoteShift (v);
      }
      else {
	configIntOutOfRange (cfg, -127, 127);
      }
    }
    else {
      configIntUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "midi.device") == 0) {
    setMIDIDevice (cfg->value);
  }

  return 0;
}

/*
 *
 */
int closeMIDI () {
  (void) close (midi_fd);
  return 0;
}

/*
 *
 */
void * MIDIInReader (void * arg) {
  int inBytes;
  unsigned char * mbp;
  unsigned char * mbpEnd;
  unsigned char B;
  int state = 0;

  /* We need to think of a way to communicate a failure back to
     the main program. */

  (void) prepareMIDI ();

  while (1) {

    if ((inBytes = read (midi_fd,
			 (void *) inBuffer,
			 (size_t) BUFFER_SIZE_BYTES)) == -1) {
      perror ("read midi_fd");
      return NULL;
    }

    if (inBytes == 0) continue;

    B = inBuffer[0];

    mbp    = inBuffer;
    mbpEnd = mbp + inBytes;

    while (mbp < mbpEnd) {
      unsigned char B = * mbp;
      unsigned char CH = B & 0x0F; /* Extract channel */
      unsigned char S = MIDI_SYSTEM_PREFIX & B;
      unsigned char status;
      unsigned char note;
      unsigned char controller;

      if ((0xF8 & B) == 0xF8) {		/* System Real-Time */
	mbp++;
	continue;
      }

      /* If we see a status byte at this point, we know that the
       * parser must begin in the idle state. Otherwise, the parser
       * is already in the idle state, or hopefully in one of the data byte
       * states, accepting further data sent using Running Status.
       */

      if (B & 0x80) {
	state = 0;
      }

      switch (state) {

      case 0:			/* Idle */

	switch (S) {		/* Determine what to do with it. */

	case MIDI_NOTEOFF:
	  state = (CH == rcvCh) ? 2 : CONSUME_TWO_BYTES;
	  break;

	case MIDI_NOTEON:
	  state = (CH == rcvCh) ? 1 : CONSUME_TWO_BYTES;
	  break;

	case MIDI_KEY_PRESSURE:
	  state = CONSUME_TWO_BYTES;
	  break;

	case MIDI_CTL_CHANGE:
	  state = (CH == rcvCh) ? 9 : CONSUME_TWO_BYTES;
	  break;

	case MIDI_PGM_CHANGE:
	  state = (CH == rcvCh) ? 8 : CONSUME_ONE_BYTE;
	  break;

	case MIDI_CHN_PRESSURE:
	  state = CONSUME_ONE_BYTE;
	  break;

	case MIDI_PITCH_BEND:
	  state = CONSUME_TWO_BYTES;
	  break;

	case 0xF0:		/* SYSEX */
	  state = CONSUME_ONE_BYTE;
	  break;

	case 0xF1:		/* Undefined */
	  break;

	case 0xF2:		/* Song position */
	  state = CONSUME_TWO_BYTES;
	  break;

	case 0xF3:		/* Song select */
	  state = CONSUME_ONE_BYTE;
	  break;

	case 0xF4:		/* Undefined */
	  break;

	case 0xF5:		/* Undefined */
	  break;

	case 0xF6:		/* Tune request */
	  break;

	case 0xF7:		/* End of SYSEX */
	  break;
	} /* switch S */

	mbp++;			/* Consume the byte. */

	break;			/* case 0 (idle) */

      case 1:			/* Note On, 2nd of 3: NOTE NUMBER */
	note = B;		/* Save the MIDI note on number */
	mbp++;
	state = 3;		/* Examine depress velocity */
	break;

      case 2:			/* Note Off 2nd of 3: NOTE NUMBER */
	note = B;		/* Save the MIDI note off number */
	mbp++;
	state = 4;		/* Examine release velocity */
	break;

      case 3:			/* Note On, 3rd of 3: VELOCITY */
	if (0 < B) {
	  oscNoteOn (note+nsh);	/* Non-zero velocity is note on */
	}
	else {
	  oscNoteOff (note+nsh); /* Note on at zero velocity is note off */
	}
	mbp++;
	state = 1;
	break;

      case 4:			/* Note Off, 3rd of 3: VELOCITY */
	oscNoteOff (note+nsh);
	mbp++;
	state = 2;
	break;

      case 5:			/* Consume one */
	mbp++;
	break;

      case 6:			/* Consume first of two */
	mbp++;
	state = 7;
	break;

      case 7:			/* Consume second of two */
	mbp++;
	state = 6;
	break;

      case 8:			/* Program change: program number */
	mbp++;
	installProgram (B);
	break;

      case 9:			/* Control change: controller */
	mbp++;
	controller = B;
	if (controller == 1) {	/* Mod wheel */
	  state = 10;
	}
	else if (controller == 11) { /* Expression pedal */
	  state = 10;
	}
	else if (controller == 64) { /* Sustain pedal */
	  state = 11;
	}
	else if (controller == 96) { /* Data increment */
	  state = 13;
	}
	else if (controller == 97) { /* Data decrement */
	  state = 14;
	}
	else {
	  state = 12;		/* Ignore controller */
	}
	break;

      case 10:			/* Mod wheel data */
	mbp++;
	set7bitGain (B);
	state = 9;
	break;

      case 11:			/* Sustain pedal data */
	mbp++;
	setRotarySusPed (B);
	state = 9;
	break;

      case 12:			/* Ignored controller data */
	mbp++;
	state = 9;
	break;

      case 13:			/* Data increment */
	mbp++;
	setScannerAdvance (1);
	state = 9;
	break;

      case 14:			/* Data decrement */
	mbp++;
	setScannerAdvance (0);
	state = 9;
	break;

      }	/* switch state */

    }
  }

  return NULL;
}
